// -*- C++ -*-
//===----------------------------------------------------------------------===//
//
// Copyright (C) 2024 Intel Corporation
//
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef _PSTL_OFFLOAD_PSTL_OFFLOAD_IMPL_H
#define _PSTL_OFFLOAD_PSTL_OFFLOAD_IMPL_H

#include "internal/usm_memory_replacement.h"
#include "internal/algorithm_redirection_impl.h"
#include "internal/memory_redirection_impl.h"
#include "internal/numeric_redirection_impl.h"

#endif // _PSTL_OFFLOAD_PSTL_OFFLOAD_IMPL_H
